use std::str::FromStr;

use fltk::{
	app,
	button::Button, 
	enums::CallbackTrigger,
	frame::Frame,
	input::*,
	prelude::*, 
	window::Window
};
use structopt::StructOpt;
use tokio::net::TcpListener;

use quic_demo::prelude::*;
use protocol::PeerId;

#[derive (Debug, StructOpt)]
struct Opt {
	#[structopt (long)]
	window_title: Option <String>,
	#[structopt (long)]
	relay_addr: Option <String>,
	#[structopt (long)]
	client_id: Option <PeerId>,
}

#[derive (Clone, Copy)]
enum Message {
	OpenPort (usize),
	ClosePort (usize),
}

fn main () -> anyhow::Result <()> {
	tracing_subscriber::fmt::init ();
	let rt = tokio::runtime::Runtime::new ()?;
	
	let opt = Opt::from_args ();
	
	let (fltk_tx, fltk_rx) = app::channel::<Message> ();
	
	let app = app::App::default ();
	let window_title = opt.window_title.clone ().unwrap_or_else (|| "PTTH client proxy".to_string ());
	let mut wind = Window::new (100, 100, 800, 600, None)
	.with_label (&window_title);
	
	let margin = 10;
	let h = 30;
	let mut x = margin;
	let mut y = margin;
	
	let mut frame_status = Frame::new (x, y, 800 - 20, h, "Forwarding 0 ports");
	
	y += h + margin;
	x = margin;
	
	{
		let w = 80;
		Frame::new (x, y, w, h, "Local port");
		x += w + margin;
		
		let w = 120;
		Frame::new (x, y, w, h, "Server ID");
		x += w + margin;
		
		let w = 80;
		Frame::new (x, y, w, h, "Server port");
		// x += w + margin;
	}
	
	y += h + margin;
	x = margin;
	
	let gui_port_0 = GuiPort::new (fltk_tx, &mut x, y, 0);
	y += h + margin;
	x = margin;
	
	let gui_port_1 = GuiPort::new (fltk_tx, &mut x, y, 1);
	y += h + margin;
	x = margin;
	
	let gui_port_2 = GuiPort::new (fltk_tx, &mut x, y, 2);
	// y += h + margin;
	// x = margin;
	
	let mut gui_ports = vec! [
		gui_port_0,
		gui_port_1,
		gui_port_2,
	];
	
	let mut forwarding_instances = vec! [
		None,
		None,
		None,
	];
	
	// y += h + margin;
	
	wind.end ();
	wind.show ();
	
	let connection_p2_p3 = rt.block_on (async move {
		let server_cert = tokio::fs::read ("quic_server.crt").await?;
		let relay_addr = opt.relay_addr.unwrap_or_else (|| String::from ("127.0.0.1:30380")).parse ()?;
		let endpoint = make_client_endpoint ("0.0.0.0:0".parse ()?, &[&server_cert])?;
		
		trace! ("Connecting to relay server");
		
		let client_id = opt.client_id.unwrap_or_else (|| "bogus_client".to_string ());
		
		let quinn::NewConnection {
			connection,
			..
		} = protocol::p2_connect_to_p3 (&endpoint, &relay_addr, &client_id).await?;
		
		Ok::<_, anyhow::Error> (connection)
	})?;
	
	while app.wait () {
		match fltk_rx.recv () {
			Some (Message::OpenPort (port_idx)) => {
				if let Ok (params) = gui_ports [port_idx].get_params () {
					let connection_p2_p3 = connection_p2_p3.clone ();
					let (shutdown_flag, shutdown_flag_rx) = tokio::sync::watch::channel (true);
					
					let task = rt.spawn (forward_port (
						connection_p2_p3,
						params,
						shutdown_flag_rx
					));
					
					forwarding_instances [port_idx].replace (ForwardingInstance {
						task,
						shutdown_flag,
					});
					
					gui_ports [port_idx].set_forwarding (true);
					
					frame_status.set_label ("Forwarding 1 port");
				}
			},
			Some (Message::ClosePort (port_idx)) => {
				if let Some (old_instance) = forwarding_instances [port_idx].take () {
					rt.block_on (async {
						old_instance.shutdown_flag.send (false)?;
						old_instance.task.await??;
						Ok::<_, anyhow::Error> (())
					})?;
				}
				
				gui_ports [port_idx].set_forwarding (false);
				frame_status.set_label ("Forwarding 0 ports");
			},
			None => (),
		}
	}
	
	Ok (())
}

fn set_active <W: WidgetExt> (w: &mut W, b: bool) {
	if b {
		w.activate ();
	}
	else {
		w.deactivate ();
	}
}

struct ForwardingInstance {
	task: tokio::task::JoinHandle <anyhow::Result <()>>,
	shutdown_flag: tokio::sync::watch::Sender <bool>,
}

async fn forward_port (
	connection_p2_p3: quinn::Connection,
	params: ForwardingParams,
	shutdown_flag_rx: tokio::sync::watch::Receiver <bool>,
) -> anyhow::Result <()>
{
	let ForwardingParams {
		client_tcp_port,
		server_id,
		server_tcp_port,
	} = params;
	
	let listener = TcpListener::bind (("127.0.0.1", client_tcp_port)).await?;
	
	trace! ("Accepting local TCP connections from P1 on {}", client_tcp_port);
	
	while *shutdown_flag_rx.borrow () {
		let mut shutdown_flag_rx_2 = shutdown_flag_rx.clone ();
		
		tokio::select! {
			x = listener.accept () => {
				let (tcp_socket, _) = x?;
				let connection = connection_p2_p3.clone ();
				let server_id = server_id.clone ();
				let shutdown_flag_rx = shutdown_flag_rx.clone ();
				
				tokio::spawn (handle_p1 (connection, server_id, server_tcp_port, tcp_socket, shutdown_flag_rx));
			},
			_ = shutdown_flag_rx_2.changed () => (),
		};
	}

	Ok::<_, anyhow::Error> (())
}

async fn handle_p1 (
	connection: quinn::Connection,
	server_id: String,
	server_tcp_port: u16,
	tcp_socket: tokio::net::TcpStream,
	shutdown_flag_rx: tokio::sync::watch::Receiver <bool>,
) -> anyhow::Result <()>
{
	let (mut local_recv, mut local_send) = tcp_socket.into_split ();
	
	debug! ("Starting PTTH connection");
	
	let (mut relay_send, mut relay_recv) = protocol::p2_connect_to_p5 (&connection, &server_id, server_tcp_port).await?;
	
	trace! ("Relaying bytes...");
	
	let task_blue = {
		let mut shutdown_flag_rx = shutdown_flag_rx.clone ();
		
		tokio::spawn (async move {
			let mut buf = vec! [0u8; 65_536];
			while *shutdown_flag_rx.borrow () {
				trace! ("Blue reading from QUIC...");
				tokio::select! {
					x = relay_recv.read (&mut buf) => {
						let bytes_read = match x? {
							None => break,
							Some (0) => break,
							Some (x) => x,
						};
						let buf_slice = &buf [0..bytes_read];
						trace! ("Uplink relaying {} bytes", bytes_read);
						local_send.write_all (buf_slice).await?;
					},
					_ = shutdown_flag_rx.changed () => (),
				};
			}
			
			debug! ("Blue QUIC --> TCP closed");
			
			Ok::<_, anyhow::Error> (())
		})
	};
	
	let task_green = {
		let mut shutdown_flag_rx = shutdown_flag_rx.clone ();
		
		tokio::spawn (async move {
			let mut buf = vec! [0u8; 65_536];
			while *shutdown_flag_rx.borrow () {
				trace! ("Green reading from TCP...");
				tokio::select! {
					x = local_recv.read (&mut buf) => {
						let bytes_read = match x? {
							0 => break,
							x => x,
						};
						let buf_slice = &buf [0..bytes_read];
						trace! ("Downlink relaying {} bytes", bytes_read);
						relay_send.write_all (buf_slice).await?;
					},
					_ = shutdown_flag_rx.changed () => (),
				};
			}
			
			debug! ("Green TCP --> QUIC closed");
			
			Ok::<_, anyhow::Error> (())
		})
	};
	
	task_blue.await??;
	task_green.await??;
	
	debug! ("Ended PTTH connection");
	
	Ok (())
}

struct GuiPort {
	input_client_port: Input,
	input_server_id: Input,
	input_server_port: Input,
	but_open: Button,
	but_close: Button,
}

struct ForwardingParams {
	client_tcp_port: u16,
	server_id: String,
	server_tcp_port: u16,
}

impl GuiPort {
	fn new (fltk_tx: fltk::app::Sender <Message>, x: &mut i32, y: i32, port_idx: usize) -> Self {
		let margin = 10;
		let h = 30;
		
		let w = 80;
		let mut input_client_port = Input::new (*x, y, w, h, "");
		*x += w + margin;
		
		let w = 120;
		let mut input_server_id = Input::new (*x, y, w, h, "");
		*x += w + margin;
		
		let w = 80;
		let mut input_server_port = Input::new (*x, y, w, h, "");
		*x += w + margin;
		
		let w = 80;
		let mut but_open = Button::new (*x, y, w, h, "Open");
		*x += w + margin;
		
		let w = 80;
		let mut but_close = Button::new (*x, y, w, h, "Close");
		// *x += w + margin;
		
		input_client_port.set_value ("5901");
		input_server_id.set_value ("bogus_server");
		input_server_port.set_value ("5900");
		
		but_open.set_trigger (CallbackTrigger::Changed);
		but_open.emit (fltk_tx, Message::OpenPort (port_idx));
		but_close.set_trigger (CallbackTrigger::Changed);
		but_close.emit (fltk_tx, Message::ClosePort (port_idx));
		
		set_active (&mut but_open, true);
		set_active (&mut but_close, false);
		
		Self {
			input_client_port,
			input_server_id,
			input_server_port,
			but_open,
			but_close,
		}
	}
	
	fn get_params (&self) -> anyhow::Result <ForwardingParams>
	{
		let client_tcp_port = u16::from_str (&self.input_client_port.value ())?;
		let server_id = self.input_server_id.value ();
		let server_tcp_port = u16::from_str (&self.input_server_port.value ())?;
		
		Ok (ForwardingParams {
			client_tcp_port,
			server_id,
			server_tcp_port,
		})
	}
	
	fn set_forwarding (&mut self, x: bool) {
		set_active (&mut self.input_client_port, !x);
		set_active (&mut self.input_server_id, !x);
		set_active (&mut self.input_server_port, !x);
		set_active (&mut self.but_open, !x);
		set_active (&mut self.but_close, x);
		
		self.but_open.set (x);
		self.but_close.set (!x);
	}
}
